#ifndef MAKEDEPEND
 #include <cstdio>
 #include <cstdlib>
 #include <map>
 #include <set>
#endif

#include "disasm.h"

static byte *data;
static size_t size;

static const bool include_disasm = false;

class flags
{
public:
	bool jump, call, end, visited;
	flags() : jump(false), call(false), end(false), visited(false) { }
};

static map<unsigned, flags> positions;
static map<unsigned, set<unsigned> > jumps;

bool haslabel(const byte *, unsigned pos)
{
	return positions[pos].jump || positions[pos].call;
}
unsigned getjumpcount(const byte *data, unsigned pos)
{
	if(!haslabel(data, pos))return 0;
	return jumps.find(pos)->second.size();
}
unsigned getjump(const byte *, unsigned pos, unsigned ind)
{
	set<unsigned>::const_iterator i;
	map<unsigned, set<unsigned> >::const_iterator j = jumps.find(pos);
	for(i=j->second.begin(); i!=j->second.end(); i++, ind--)
		if(!ind)return *i;
	return 0;
}
bool alreadycompiled(const byte *, unsigned pos)
{
	map<unsigned, flags>::iterator i = positions.find(pos);
	return i->second.visited;
}
void MarkVisited(const byte *, unsigned pos)
{
	positions.find(pos)->second.visited = true;
}
void RequestCompile(const byte *data, unsigned pos, bool enterblock, bool doendjump)
{
	unsigned endjump=0;
	map<unsigned, flags>::iterator i = positions.find(pos);
//	printf("Finding at %X (%X)\n", i->first, pos);
	while(!i->second.end)
	{
//		printf("Will compile %X\n", i->first);
//		i->second.visited = true;
		++i;
		if(i->second.visited)
		{
//			printf("Will NOT compile %X\n", i->first);
			if(doendjump)endjump = i->first;
			--i;
			break;
		}
	}
	compile(data, pos, i->first+1 - pos, enterblock, endjump);
	printf("\n\n");
	i->second.visited = true;
}

int main(int argc, const char *const *argv)
{
	printf("// Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n\n");
	
	if(argc != 2)
	{
		printf(
			"Usage: dis <filename.gb> | less -Sr\n"
			"\n"
			"The program will disassemble at least the first two 16k pages of the\n"
			"ROM image and translate them to some kind of pseudo-C language.\n");
		return 0;			
	}
	FILE *fp = fopen(argv[1], "rb");
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	rewind(fp);
	
	data = new byte[size];
	fread(data, 1, size, fp);
	fclose(fp);
	
	unsigned p = 0x100;
	
	map<unsigned, flags>::iterator i;
	
	if(include_disasm)
		printf("// This all you will get:\n\n");
	
	for(;;)
	{
		bool jump, call, end=false;
		unsigned nextpos;
		unsigned len = DisAsm(data, p, include_disasm,
		                      &jump, &call, &end,
		                      &nextpos);
		                 
		if(include_disasm)printf("\n");
		
		if(p==0x100)
			positions[p].call = true;

		positions[p].end = end;
		positions[p].visited = true;
		
		if(call)positions[nextpos].call = true;
		if(jump)positions[nextpos].jump = true;
		
		if(call || jump)
		{
			jumps[nextpos].insert(p);
		}
		
		if(end)
		{
End:		for(i=positions.begin(); i != positions.end(); ++i)
				if(!i->second.visited)
				{
					p = i->first;
					goto Found;
				}
			break;
Found:;
		}
		else
		{
			p += len;
			if(positions[p].visited)goto End;
		}
	}
	
	printf("\n// The decompilation:\n\n");
	
	for(i=positions.begin(); i != positions.end(); ++i)i->second.visited = false;
	
	printf("*** Main level:\n");
	RequestCompile(data, 0x100, true);

	for(i=positions.begin(); i != positions.end(); ++i)
		if((i->second.call || i->second.jump)
		&& !i->second.visited)
		{
			printf("*** Main level:\n");
			RequestCompile(data, i->first, true);
		}
			
	return 0;
}
