// This is part of emulator.cc of GameBoy emulator.
// Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)

#include "timer.h"

static class tmareg
{
	byte value;
public:
	tmareg() : value(0)
	{
	}
	inline byte read() const
	{
		return value;
	}
	inline void write(const byte data)
	{
		value = data;
	}	
} TMA;

static void TimaUpdater();

static class timareg
{
	bool enabled;
	byte value;
	timer tim;
public:
	timareg() : enabled(true), value(0), tim(16384, TimaUpdater)
	{
	}
	inline byte read() const
	{
		return value;
	}
	inline void write(const byte data)
	{
		value = data;
	}
	inline void update()
	{
		if(enabled)
			if(!++value)
			{
				value=TMA.read();
				IRQ(4); // 1=vblank, 2=lcd, 4=timer, 8=serial, 16=joy
			}
	}
	inline void setenabled(bool v)
	{
		enabled = v;
	}
	inline void setspeed(unsigned n)
	{
		tim.changerate(n);
	}
} TIMA;

static void TimaUpdater()
{
	TIMA.update();
}

static class tacreg
{
public:
	inline void write(const byte data)
	{
		TIMA.setenabled(data&4);
		switch(data&3)
		{
			case 1: TIMA.setspeed(262144); break;
			case 2: TIMA.setspeed(65536); break;
			case 3: TIMA.setspeed(16384); break;
			case 0: TIMA.setspeed(4096); break;
		}
	}
} TAC;
